/*     This file is part of KFireSaver.
 *
 *     KFireSaver is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     KFireSaver is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with KFireSaver; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *     Authors:
 *       David Sansome <me@davidsansome.com>
 *       KoRaL <asy@libero.it>
 *
 *     (see the changelog for more information)
 */

#include <qgl.h>
#include <math.h>
#include <kstddirs.h>
#include <klocale.h>
#include <qdatetime.h>
#include <qmessagebox.h>
#include <stdlib.h>
#include <kconfig.h>
#include "kfiresaver.h"
#include "particle.h"

#define MSECPERIOD 10.0f
#define PERIOD MSECPERIOD/1000.0f

KFireSaver::KFireSaver( QWidget *parent, const char *name )
	: QGLWidget( parent, name )
{
	readConfig();

	particleList.setAutoDelete(true);

	kdeLogo.setAlphaBuffer(true);
	kdeLogo.load(locate("data","kfiresaver/kde.png"));
	tux.setAlphaBuffer(true);
	tux.load(locate("data","kfiresaver/tux.png"));

	QTime midnight( 0, 0, 0 );
	srand( midnight.secsTo(QTime::currentTime()) );

	if (enableBottomFire)
	{
		for (int i=0 ; i<300 ; i++)
		{
			Particle* particle = new Particle;
			particle->setValues(Particle::FireParticle,bottomFireColour);
			particleList.append(particle);
		}
	}

	startTimer(MSECPERIOD);
}

void KFireSaver::initializeGL()
{
	/*static GLfloat pos[4] = {5.0, 5.0, -1.0, 1.0 };
	glLightfv( GL_LIGHT0, GL_POSITION, pos );
	glEnable( GL_CULL_FACE );
	glEnable( GL_LIGHTING );
	glEnable( GL_LIGHT0 );*/
	glDisable( GL_DEPTH_TEST );
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);

	glShadeModel( GL_SMOOTH );
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glHint( GL_POINT_SMOOTH_HINT, GL_NICEST );

	// Texture

	QImage buf;

	if (!buf.load(locate("data","kfiresaver/particle.bmp")))
	{
		qWarning("Could not read image file, using single-color instead.");
		QImage dummy(0,0,0);
		dummy.fill( Qt::white.rgb() );
		buf = dummy;
	}
	particleTexture = QGLWidget::convertToGLFormat(buf);  // flipped 32bit RGBA

	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glEnable( GL_TEXTURE_2D );

	//glEnable( GL_NORMALIZE );
}


void KFireSaver::resizeGL( int width, int height )
{
//	GLfloat w = /*(float) width / (float) height*/ 1.0f;
//	GLfloat h = 1.0;

    glViewport( 0, 0, width, height );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    /*glFrustum( -w, w, -h, h, 5.0, 60.0 );*/
    glOrtho( -10.0f, 10.0f, -10.0f, 10.0f, 5.0, 60.0 );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef( 0.0, 0.0, -40.0 );
	/*glOrtho(0,640,0,480,-1,1);*/
}

void KFireSaver::paintGL()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glLoadIdentity();

	glTexImage2D( GL_TEXTURE_2D, 0, 3, particleTexture.width(), particleTexture.height(), 0,
		  GL_RGBA, GL_UNSIGNED_BYTE, particleTexture.bits() );

	GLfloat unitx = xToGl(1),
		unity = yToGl(1);

	for (int i=0 ; i<(int)particleList.count() ; i++)
	{
		Particle* particle = particleList.at(i);

		glLoadIdentity();
		//perspective projection
		GLfloat pposx = particle->xpos * 256 / (256 + 2*particle->ypos),
			pposy = particle->zpos * 256 / (256 + 2*particle->ypos);
		glTranslatef( pposx, pposy, -50.0f);
		GLfloat sfactor = 256 + 15 * particle->ypos;
		if (sfactor < 10.0)
		    sfactor = 10.0;
		if (sfactor > 512.0)
		    sfactor = 512.0;
		GLfloat size = particle->pixelSize * 256 / sfactor;
		glScalef( size, size, size );
		//bright changes with the following curve: "2*k - k^2" (or "k(2-k)")
		GLfloat alpha = particle->life / particle->startLife;
		if (particle->useConstantAlpha)
			alpha = particle->constantAlpha;
		if (particle->flicker < 0) {
			alpha = 0.05;
			if (++particle->flicker >= 0)
				particle->flicker = FLICKER_FRAMES_DELAY;
		} else if (particle->flicker > 0) {
			if (--particle->flicker <= 0)
				particle->flicker = -FLICKER_FRAMES_DELAY;
		}
		glColor4f(particle->colour[0],particle->colour[1],particle->colour[2], alpha * (2 - alpha));
		glBegin( GL_TRIANGLE_STRIP );
			/* Top Right */
			glTexCoord2d( 1, 1 );
			glVertex2f( unitx, unity );
			/* Top Left */
			glTexCoord2d( 0, 1 );
			glVertex2f( -unitx, unity );
			/* Bottom Right */
			glTexCoord2d( 1, 0 );
			glVertex2f( unitx, -unity );
			/* Bottom Left */
			glTexCoord2d( 0, 0 );
			glVertex2f( -unitx, -unity );
		glEnd();

		if (sfactor == 10.0 || sfactor == 512.0) {
		    particleList.remove(i--);
		    continue;
		}

		particle->applyPhysics( PERIOD );

		switch (particle->particleType)
		{
		case Particle::FireParticle:
			if (particle->life <= 0.0f)
				particle->setValues(Particle::FireParticle,bottomFireColour);
			break;
		case Particle::FireWorkLeaderParticle:
			if (particle->zspeed <= 0.0f)
			{
				explodeFirework(particle);
				particleList.remove(i--);
				//light of explosion (must be configurable) - it is now!
				if (enableFlash)
				{
					glClearColor( 0.5, 0.5, 0.5, 0.0 );
					glClear( GL_COLOR_BUFFER_BIT );
					glClearColor( 0.0, 0.0, 0.0, 0.0 );
				}
			}
			else if (enableTrails)
			{
				int random = (int) (DRAND * 2.0f);
				if (random == 1)
				{
					Particle* wake = new Particle;
					wake->setValues(Particle::FireWorkWakeParticle,0,particle);
					particleList.append(wake);
				}
			}
			break;
		case Particle::FireWorkDebrisParticle:
		/*{
			int random = (int) (DRAND * 50.0f);
			if (random == 1)
			{
				Particle* wake = new Particle;
				wake->setValues(Particle::FireWorkWakeParticle,0,particle);
				particleList.append(wake);
			}
		}*/
		case Particle::FireWorkWakeParticle:
		case Particle::KdeLogoParticle:
			/*if ((particle->xpos <= -13.35f) || (particle->xpos >= 13.35f))
			{
				//particle->xpos = 0.0f;
				particle->xspeed -= particle->xspeed*2;
			}*/
			if (particle->life <= 0.0f || particle->zpos < -10.0 ||
			    particle->xpos < -15.0 || particle->xpos > 15.0 ||
			    particle->ypos > 15.0)
				particleList.remove(i--);
			break;
		}
	}

	int random = (int) (DRAND * 10.0 * (float)fireworksFrequency);
	if (random == 1)
	{
		Particle* particle = new Particle;
		particle->setValues(Particle::FireWorkLeaderParticle);
		particleList.append(particle);
	}
	random = (int) (DRAND * 200.0 * (float)logoFrequency);
	if (random == 1)
	{
		if ((enableTux) && (enableKdeLogo))
			random = (int) (DRAND * 2.0);
		else if (enableTux)
			random = 1;
		else if (enableKdeLogo)
			random = 0;
		else
			return;
		//light of explosion
		if (enableFlash)
		{
			glClearColor( 1.0, 1.0, 1.0, 0.0 );
			glClear( GL_COLOR_BUFFER_BIT );
			glClearColor( 0.0, 0.0, 0.0, 1.0 );
		}
		if (random == 1)
			burnLogo(tux);
		else
			burnLogo(kdeLogo);
	}

	//glPopMatrix();
}

void KFireSaver::explodeFirework(Particle* fireWorkLeaderParticle)
{
	int number = 100 + (int) (DRAND * 500.0),
	    cscheme = selectColourScheme(),
	    cscheme2 = selectColourScheme();
	GLfloat power = DRAND * 10.0 + 2.0,
		powermin = DRAND * power;
	GLfloat displace[3] = {0.0,0.0,0.0};
//	displace[0] = power * (DRAND - 0.5) / 2;
//	displace[1] = power * (DRAND - 0.5) / 2;
//	displace[2] = power * (DRAND) / 2;
	bool bicolor = DRAND > 0.9;	// 1/10 is bicolored
	if (!enableCombos)
		bicolor = false;
	bool flickers = DRAND > 0.9;	// 1/10 flickers
	if (!enableSparkle)
		flickers = false;
	if (cscheme == 1) {	//flickering stars
		bicolor = false;
		flickers = DRAND > 0.1;	// 9/10 stars flickers
		powermin = power / 3;
		displace[2] = power;
	}
	for (int i=0 ; i<number ; i++)
	{
		Particle* particle = new Particle;
		particle->setValues(Particle::FireWorkDebrisParticle,
			bicolor && (i>number/2) ? cscheme2 : cscheme,
			fireWorkLeaderParticle, powermin, power,
			flickers, displace );
		if (cscheme == 1) {
			particle->zacc = -9.807 * (0.25 + DRAND/4);
			particle->zspeed *= 0.7;
			particle->life *= 2;
		}
		particleList.append(particle);
	}
}

int KFireSaver::selectColourScheme()
{
	for(;;)
	{
		int c = DRAND * 7.0;
		switch(c)
		{
		case 0: if (enableRed) return c; break;
		case 1: if (enableOrange) return c; break;
		case 2: if (enableGreen) return c; break;
		case 3: if (enableBlue) return c; break;
		case 4: if (enableWhite) return c; break;
		case 5: if (enablePurple) return c; break;
		case 6: if (enableDeepGreen) return c; break;
		}
	}
}

void KFireSaver::readConfig()
{
	KConfig config("kfiresaverrc",true,false);

	enableFlash = config.readBoolEntry("enableFlash",false);
	enableSparkle = config.readBoolEntry("enableSparkle",true);

	enableRed = config.readBoolEntry("enableRed",true);
	enableOrange = config.readBoolEntry("enableOrange",true);
	enableGreen = config.readBoolEntry("enableGreen",true);
	enableBlue = config.readBoolEntry("enableBlue",true);
	enableWhite = config.readBoolEntry("enableWhite",true);
	enablePurple = config.readBoolEntry("enablePurple",true);
	enableDeepGreen = config.readBoolEntry("enableDeepGreen",true);
	enableCombos = config.readBoolEntry("enableCombos",true);

	fireworksFrequency = 11 - config.readNumEntry("fireworksFrequency",5);

	enableBottomFire = config.readBoolEntry("enableBottomFire",true);
	bottomFireColour = config.readNumEntry("bottomFireColour",1);

	enableTux = config.readBoolEntry("enableTux",true);
	enableKdeLogo = config.readBoolEntry("enableKdeLogo",false);

	logoFrequency = 11 - config.readNumEntry("logoFrequency",5);
	
	enableTrails = config.readBoolEntry("enableTrails",true);
}

void KFireSaver::timerEvent(QTimerEvent*)
{
	updateGL();
}

void KFireSaver::burnLogo(QImage image)
{
	for (int y=0 ; y<image.height() ; y+=2)
	{
		for (int x=0 ; x<image.width() ; x+=2)
		{
			if ((qBlue(image.pixel(x,y)) == 255) && (qRed(image.pixel(x,y)) == 0) && (qGreen(image.pixel(x,y)) == 0))
				continue;

			Particle* particle = new Particle;
			particle->xpos = xToGl(x) + DRAND * 0.2f - 0.1f;
			particle->zpos = yToGl(y) + DRAND * 0.2f + 2.0f;
			particle->xspeed = 0.1 * (float)(x-(image.width()/2));
			particle->zspeed = 0.1 * (float)(y-(image.height()/2));
			particle->ypos = 0;
			particle->yspeed = DRAND * 20.0f - 10.0f;
			particle->xacc = particle->yacc = 0;
			particle->zacc = -9.807f / 5.0f;
			particle->colour[0] = qRed(image.pixel(x,y)) / 255.0f;
			particle->colour[1] = qGreen(image.pixel(x,y)) / 255.0f;
			particle->colour[2] = qBlue(image.pixel(x,y)) / 255.0f;
			particle->useLife = true;
			particle->startLife = 1.5f;
			particle->life = 1.5f;
			particle->pixelSize = 3;
			particle->particleType = Particle::KdeLogoParticle;
			particleList.append(particle);
		}
	}
}

GLfloat KFireSaver::xToGl(int x)
{
	return ((float)x) / ((float)width()) * 20.0f;
}

GLfloat KFireSaver::yToGl(int y)
{
	return ((float)y) / ((float)height()) * 20.0f;
}
