/*     This file is part of KFireSaver.
 *
 *     KFireSaver is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     KFireSaver is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with KFireSaver; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *     Authors:
 *       David Sansome <me@davidsansome.com>
 *       KoRaL <asy@libero.it>
 *
 *     (see the changelog for more information)
 */

#include "particle.h"
#include "kfiresaver.h"
#include <math.h>
#include <stdlib.h>

void Particle::setValues(ParticleType pt, int cs, Particle* leader,
	GLfloat powermin, GLfloat powermax, bool flickers, GLfloat *displace)
{
	//                                red     green white deep-green
	//                                   orange  blue  purple
	const GLfloat c_red_min[10] =   { 0, 0.9, 0, 0, 1, 0, 0 };
	const GLfloat c_red_max[10] =   { 1,   1, 0, 0, 1, 1, 0 };
	const GLfloat c_green_min[10] = { 0, 0.4, 0, 0, 1, 0, 0 };
	const GLfloat c_green_max[10] = { 0, 0.5, 1, 0, 1, 0, 1 };
	const GLfloat c_blue_min[10] =  { 0, 0.1, 0, 0, 1, 0, 0 };
	const GLfloat c_blue_max[10] =  { 0, 0.2, 0, 1, 1, 1, 1 };
	
	//                                 blue      purple
	//                                      orange
	const GLfloat bf_red_min[10] =   {   0, 0.9, 0 };
	const GLfloat bf_red_max[10] =   {   0,   1, 1 };
	const GLfloat bf_green_min[10] = {   0, 0.4, 0 };
	const GLfloat bf_green_max[10] = { 0.6, 0.5, 0 };
	const GLfloat bf_blue_min[10] =  {   1, 0.1, 0 };
	const GLfloat bf_blue_max[10] =  {   1, 0.2, 1 };

	particleType = pt;

	xacc = yacc = 0;
	flicker = 0;

	switch (particleType)
	{
	case FireParticle:
		//born in a 20*10 rectangle centered in 0,0,-9.9
		//speed 3-5 [m/s] blue/cyan colored

		zpos = -9.9;
		zspeed = DRAND * 2.0 + 5.0;
		zacc = -9.807;

		xpos = DRAND * 20.0 - 10.0;
		xspeed = DRAND * 2.0 - 1.0;	// -1 .. 1

		ypos = /*DRAND * 10.0 - 5.0*/-10.0f;
		yspeed = DRAND * 1.0 + 0.5;

		colour[0] = bf_red_min[cs] + (bf_red_max[cs]-bf_red_min[cs]) * DRAND;
		colour[1] = bf_green_min[cs] + (bf_green_max[cs]-bf_green_min[cs]) * DRAND;
		colour[2] = bf_blue_min[cs] + (bf_blue_max[cs]-bf_blue_min[cs]) * DRAND;
		useLife = true;
		startLife = life = DRAND * 0.5f + 0.25f;
		/*useConstantAlpha = true;
		constantAlpha = 0.05f;*/
		useConstantAlpha = false;
		pixelSize = 15.0f;
		
		antiGravity = true;
		
		/*zspeed *= 4.0;
		zpos = 0.0;*/

		break;
	case FireWorkLeaderParticle:
		zpos = -9.9f;
		zspeed = DRAND * 5.0f + 14.0f;
		zacc = -9.807;

		xpos = DRAND * 8.0f - 4.0f;
		xspeed = (DRAND * 8.0f - 4.0f + xpos) / 2.0f;

		ypos = DRAND * 8.0f - 4.0f;
		yspeed = DRAND * 6.0f - 3.0f;

		colour[0] = 1.0f;
		colour[1] = DRAND * 0.6f;
		colour[2] = 0.0f;

		useLife = false;
		useConstantAlpha = false;
		startLife = life = 1.0f;
		pixelSize = 5;
		antiGravity = false;
		// Leaders don't use life - they explode when their zspeed drops to 0
		break;
	case FireWorkWakeParticle:
		xpos = leader->xpos;
		ypos = leader->ypos;
		zpos = leader->zpos;

		xspeed = 0 + DRAND * 2.0f - 1.0f;
		yspeed = 0;
		zspeed = 0.5f;
		
		colour[0] = leader->colour[0];
		colour[1] = leader->colour[1];
		colour[2] = leader->colour[2];
		
		useLife = true;
		startLife = life = leader->life;
		pixelSize = leader->pixelSize * 2;

		useConstantAlpha = false;

		break;
	case KdeLogoParticle:
		break;
	case FireWorkDebrisParticle:
		//explosion
		xpos = leader->xpos;
		ypos = leader->ypos;
		zpos = leader->zpos;

		float module = powermin + DRAND * (powermax - powermin),
		      theta = DRAND * M_PI * 2.0,
		      u = DRAND * 2.0 - 1.0,
		      root = sqrt( 1 - u*u );
		xspeed = 2*leader->xspeed + module * root * cos(theta) * (1 + DRAND/3);
		yspeed = 2*leader->yspeed + module * root * sin(theta) * (1 + DRAND/3);
		zspeed = module * u * (0.9 + DRAND/3);

		if (displace != NULL) {
			xspeed += displace[0];
			yspeed += displace[1];
			zspeed += displace[2];
		}

		colour[0] = c_red_min[cs] + (c_red_max[cs]-c_red_min[cs]) * DRAND;
		colour[1] = c_green_min[cs] + (c_green_max[cs]-c_green_min[cs]) * DRAND;
		colour[2] = c_blue_min[cs] + (c_blue_max[cs]-c_blue_min[cs]) * DRAND;

		float size = DRAND * 2.0 + 2.0;
		zacc = -9.807 / (6 - size);

		useLife = true;
		useConstantAlpha = false;
		startLife = life = size / 2;	//old value: 2.0f
		pixelSize = size;		//old value: 3.0f

		flicker = !flickers ? 0 :
			(int) ((DRAND * 2.0 - 1.0) * (float)FLICKER_FRAMES_DELAY);
		
		antiGravity = false;

		break;
	}
}

void Particle::applyPhysics( GLfloat deltaTime )
{
	if (useLife)
	    life -= deltaTime;
	xpos += xspeed * deltaTime;
	ypos += yspeed * deltaTime;
	zpos += zspeed * deltaTime;
// n/a	xspeed += xacc * deltaTime;
// n/a	yspeed += yacc * deltaTime;
//hack: air friction
	xspeed *= 0.995;
	if (!antiGravity)
	{
		yspeed *= 0.995;
		zspeed += zacc * deltaTime;
	}
}
